//           ___   ____               ___       ___
// \ \  / / | _   |  __| \ \  / / || | __  || || _ |
//  \ \/ /  |___  | |__   \ \/ /  || |___  || ||___|
//   \  /   | _   | _  |   \  /   ||  __ | || ||\\
//    \/    |___  |___ |    \/    || ____| || || \\
//
// Copyright (c) 2021 Piotr Biernat. https://pbiernat.dev. MIT License
// Repo: https://git.pbiernat.dev/golang/vegvisir

package server

import (
	"context"
	"fmt"
	"log"
	"os"
	"os/signal"
	"regexp"
	"strings"
	"time"
	"vegvisir/pkg/cache"
	"vegvisir/pkg/config"

	"github.com/valyala/fasthttp"
)

const (
	Version = "0.1"
	Name    = "Vegvisir/" + Version
)

type Server struct {
	Config config.Config

	cFilePath string
	rCache    map[string]cache.RouteCache // Internal route cache
	routeCM   cache.RouteCacheManager     // Redis route cache
	respCM    cache.ResponseCacheManager  // Redis response cache
}

func NewServer(cPath string) *Server {
	datastore := cache.NewRedisDatastore("127.0.0.7", 6379) // FIXME use config or env...

	return &Server{
		cFilePath: cPath,
		rCache:    make(map[string]cache.RouteCache),
		routeCM:   cache.NewRouteCacheManager(datastore, 30),    //FIXME use ttl(seconds) from config or env...
		respCM:    cache.NewResponseCacheManager(datastore, 30), //FIXME use ttl(seconds) from config or env...
	}
}

func (s *Server) Run() {
	if err := s.Config.Load(s.cFilePath); err != nil {
		log.Fatalln("Unable to find config file: ", s.cFilePath, err)
	}

	go func() {
		serverAddress := s.Config.Server.Address + ":" + fmt.Sprint(s.Config.Server.Port)
		if err := fasthttp.ListenAndServe(serverAddress, s.mainHandler); err != nil {
			log.Fatalf("Server panic! Error message: %s", err)
		}
	}()

	log.Println("Server started")

	// Wait for an interrupt
	interrupt := make(chan os.Signal, 1)
	signal.Notify(interrupt, os.Interrupt, os.Kill)
	<-interrupt

	log.Println("SIGKILL or SIGINT caught, shutting down...")

	// Attempt a graceful shutdown
	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()

	s.Shutdown(ctx)
	log.Println("Server shutdown successfully.")
}

func (s *Server) Shutdown(ctx context.Context) { // TODO: wait for all connections to finish
	log.Println("Shuting down finished")
}

func (s *Server) mainHandler(ctx *fasthttp.RequestCtx) {
	ctx.Response.Header.Add(fasthttp.HeaderServer, Name)

	// move all below logic to concrete handler or sth....
	reqUri, sReqUri, sReqMethod := ctx.RequestURI(), string(ctx.RequestURI()), string(ctx.Method())
	log.Println("Incomming request:", sReqMethod, sReqUri)

	found, route := s.findRouteByRequestURI(reqUri)
	if !found {
		// FIXME: return 404/5xx error in repsonse, maybe define it in Backend config?
		ctx.SetStatusCode(fasthttp.StatusNotFound)
		return
	}

	// handle response caching
	if ok, data := s.respCM.Load(sReqUri); ok {
		log.Println("Read resp from cache: ", route.TargetUrl)

		ctx.SetBody([]byte(data.Body))
	} else {
		log.Println("Send req to backend url: ", route.TargetUrl)

		// prepare to send request to backend - separate
		bckReq := fasthttp.AcquireRequest()
		bckResp := fasthttp.AcquireResponse()
		defer fasthttp.ReleaseRequest(bckReq)
		defer fasthttp.ReleaseResponse(bckResp)

		// copy headers from backend response and prepare request for backend - separate
		bckReq.SetRequestURI(route.TargetUrl)
		bckReq.Header.SetMethod(sReqMethod)

		err := fasthttp.Do(bckReq, bckResp)
		if err != nil {
			ctx.SetStatusCode(fasthttp.StatusInternalServerError)
			return
		}

		ctx.Response.Header.SetBytesV(fasthttp.HeaderContentType, bckResp.Header.ContentType())
		ctx.SetStatusCode(bckResp.StatusCode())
		ctx.SetBody(bckResp.Body())

		// save response to cache
		respCache := cache.ResponseCache{
			URL:  sReqUri,
			Body: string(bckResp.Body()),
			// Headers: []
		} // FIXME: prepare resp cache struct in respCM.Save method or other service...
		s.respCM.Save(sReqUri, respCache)
	}
}

func (s *Server) findRouteByRequestURI(uri []byte) (bool, cache.RouteCache) {
	var sUri string = string(uri)

	for bId := range s.Config.Backends {
		bck := s.Config.Backends[bId]
		if !strings.Contains(sUri, bck.PrefixUrl) {
			continue
		}

		for rId := range bck.Routes {
			route := &bck.Routes[rId]

			// if ok, cRoute := s.rCacheManager.Load(sUri); ok {
			// 	return true, cRoute
			// }
			if cRoute, ok := s.rCache[sUri]; ok {
				return true, cRoute
			}

			rgxp := regexp.MustCompile(fmt.Sprintf("%s%s", bck.PrefixUrl, route.Pattern))
			if rgxp.Match(uri) {
				targetUrl := bck.BackendAddress + rgxp.ReplaceAllString(sUri, route.Target)

				cRoute := cache.RouteCache{ // FIXME: data duplication and use short alias for backend and route!
					// Backend:   bck,
					// Route:     *route,
					SourceUrl: sUri,
					TargetUrl: targetUrl,
				}

				// s.rCacheManager.Save(sUri, cRoute)
				s.rCache[sUri] = cRoute

				return true, cRoute
			}
		}
	}

	return false, cache.RouteCache{}
}
